# -*- coding: utf-8 -*-
from PyQt5 import QtCore, QtGui, QtWidgets

from API import BuilderPanelPlugin, StashItem
from NL.GUI.Presenter.Presenter import Presenter
from NL.CommonConcepts.Configurations.ConfigurationCopy import configurationCopy
from NL.CommonConcepts.Configurations.Utilities import center

from QuantumATK import Angstrom, BulkConfiguration
import numpy, math
import copy

from .UI.TubeWrapperWidget import TubeWrapperWidget

def displacement(x, ycell, xcell, wrapping_angle, rise_z):
    """
    Function for converting a sheet coordinate into a partly wrapped tube
    @param x              : Coordinates of atom
    @param ycell          : Cell length in Y = full circumference of tube
    @param xcell          : Cell length in X
    @param wrapping_angle : Wrapping angle of the tube in radians
    """
    # Calculate the radius of the incomplete wrapped tube
    # The second term accounts for spread of atoms in X, like in MoS2
    # (else we map all atoms to the same radius regardless of X)
    radius = ycell/wrapping_angle + (x[0]-xcell/2.)
    # Find the angle of the current atom
    # First part is basically the fractional y coordinate in the interval -0.5,0.5
    # and thus we map the fractional y coordinate [0..1] onto [-wrapping_angle/2,wrapping_angle/2]
    angle = (x[1]-ycell/2.)/ycell*wrapping_angle
    # Rise in Z for helical structures
    dz = rise_z * x[1]/ycell
    # Return atom position of the wrapped atom
    return numpy.array([ radius*math.cos(angle),
                         radius*math.sin(angle),
                         x[2] + dz ])

class TubeWrapperPlugin(BuilderPanelPlugin):
    """
    Plugin for "wrapping" a periodic structure into a (possibly open) tube
    """
    def __init__(self, parent=None):
        """
        Constructor.

        @parent : The widget that holds the plugin.
        """
        self._widget = TubeWrapperWidget()

        BuilderPanelPlugin.__init__(self, parent)

        self._widget._angle_edit.setBottom(0.)
        self._widget._angle_edit.setTop(360.)
        self._widget._rise_z_edit.setSuffix(u'Å')
        self.reset()

        # Setup signals
        self._widget._apply_button.clicked.connect(self.applyClicked)
        self._widget._preview_button.clicked.connect(self.updateAndShowPreview)
        self._widget._angle_slider.valueChanged[int].connect(self.updateAndShowPreview)
        self._widget._angle_edit.editingFinished.connect(self.updateAndShowPreview)
        self._widget._tips_button.clicked.connect(self.tipsButtonPressed)

        # Setup a how-to page.
        self._tips_page = QtWidgets.QTextEdit()
        self._tips_page.setHtml(HTML_CODE)
        self._tips_page.setLineWrapColumnOrWidth(40)
        self._tips_page.setReadOnly(True)
        self._tips_page.setFontPointSize(14)
        self._tips_page.setMinimumHeight(475)
        self._tips_page.setWindowIcon(QtGui.QIcon(':/vnl_icon.png'))
        font = self._tips_page.font()
        font_metrics = QtGui.QFontMetrics(font)
        self._tips_page.setFixedWidth(font_metrics.width('x'*50))
        self._tips_page.setWindowTitle("Tips on using the TubeWrapper")
        self._tips_page.setWindowFlags(QtCore.Qt.Tool)
        # Resize the how-to button.
        self._widget._tips_button.setFixedWidth(font_metrics.width('   Tips   '))

        # Initialize the preview member data to exist
        self._preview = None

        # Protection from callbacks on changing widget values programmatically
        self._setting = False

        # Protection from callbacks when changing the configuring on Apply
        self._applying = False

    def title(self):
        """
        @return The title of the panel.
        """
        return self.tr("Tube Wrapper")

    def widget(self):
        """
        @return The widget presented to the user.
        """
        return self._widget

    @staticmethod
    def group():
        return 'Coordinate Tools'

    def showPreview(self):
        """
        Responsible for showing the preview
        """
        self.stash().setPreview(self._preview, self)

    def tipsButtonPressed(self):
        """
        Show the tips page if it is not already shown
        """
        # If it is not already visible, show it.
        if not(self._tips_page.isVisible()):
            self._tips_page.show()
        # Otherwise hide it.
        else:
            self._tips_page.hide()

    def applyTo(self, presenter):
        """
        Apply the requested wrapping to a specific presenter
        (or rather, the configuration on the presenter)
        @param presenter : The presenter to apply to
        """
        # Get wrapping angle and helical rise from the widget
        angle = self._widget._angle_edit.value()
        rise_z = self._widget._rise_z_edit.value()

        if angle == 0.:
            # No wrapping - can't use displacement because it divides by the angle
            # But coordinates are trivially unchanged anyway
            new_coordinates = numpy.array(self._c)
        else:
            # Define a function of one variable, f(c), for displacing the atoms
            f = lambda c : displacement(c,self._full_radius,self._x_cell,angle*math.pi/180, rise_z)
            # Apply the function to find new displaced coordinates
            new_coordinates = numpy.apply_along_axis(f,1,self._c)

        # Update coordinates, keeping the tube centered
        presenter.changeAtoms(indices=None,positions=new_coordinates*Angstrom)
        presenter.changeCoordinates(center, axes=(True,True,False))

    def updateAndShowPreview(self):
        """
        Called whenever something changes, to show the new state as a preview
        """
        widget = self._widget
        # How did we get here? If setting, get out
        if self._setting:
            return
        self._setting = True

        # Who called us? Update the corresponding partner widget
        if widget.sender() == widget._angle_slider:
            widget._angle_edit.setValue(widget._angle_slider.value())
        elif widget.sender() == widget._angle_edit:
            widget._angle_slider.setValue(widget._angle_edit.value())
        self._setting = False

        # Apply wrapping to the preview item and show the preview
        self.applyTo(self._preview)
        self.showPreview()

    def applyClicked(self):
        """
        Called when the 'Apply' button is pressed.
        """
        self._applying = True

        # Set Undo mark on active presenter
        presenter = self.stash().activePresenter()
        presenter.setUndoMark('Tube Wrapper')

        # Apply the stretch to the real configuration
        self.applyTo(presenter)

        # When done, remove the preview
        self.reset()
        self._applying = False
        self.stash().setPreview(None)

    def reset(self):
        """
        Reset angle to 360
        """
        self._widget._angle_edit.setValue(360.)
        self._widget._angle_slider.setValue(360.)
        self._widget._rise_z_edit.setValue(0.)

    def activeConfigurationModified(self):
        """
        Listens to changes on the stash
        """
        if self._applying:
            return

        configuration = configurationCopy(self.stash().activeConfiguration())

        # This plugin is only enabled for bulk configurations
        if isinstance(configuration, BulkConfiguration) and (len(configuration) > 0):
            self.setAvailable(True)

            # Make a local copy of the configuration to use for the preview
            if self._preview == None:
                self._preview = Presenter(configuration)
            else:
                self._preview.setConfiguration(configuration)

            # Centering in X so atoms in the middle of cell lie on the main radius of the tube
            # This changes "configuration" by reference but this is a local configurationCopy
            self._preview.changeCoordinates(center, axes=(True,False,False))

            # This statement is crucial - we need a copy, not a reference, to the ORIGINAL CENTERED coordinates
            self._c = configuration.cartesianCoordinates().inUnitsOf(Angstrom)*Angstrom

            # Store the radius and cell size in X
            self._full_radius = numpy.linalg.norm(configuration._cell().inUnitsOf(Angstrom)[1])
            self._x_cell = numpy.linalg.norm(configuration._cell().inUnitsOf(Angstrom)[0])

        else:
            self.setAvailable(False)

HTML_CODE = """<html>
<head>
  <meta http-equiv="content-type" content="text/html; charset=ISO-8859-1">
</head>
<body bgcolor="#FFFFFF" text="#000000">
<p>Specify an angle to wrap a periodic system onto a circular arc, or a tube if the angle is 360 degrees.
The original <b>B (Y)</b> axis will be mapped onto the arc, while the tube axis will lie along <b>C (Z)</b>.
If the helical rise is non-zero, a helix is created such that the difference in Z coordinate between the
beginning and end is the specified "helical rise".</p>
<br/>
<p><b>Tips:</b></p>
<ul>
<li>The original cell should be orthorhombic for proper results.</li>
<li>If your cell is not aligned in the way described above, use Bulk Tools&gt;Swap Axes to align the axes as needed.</li>
<li>Make sure the cell size in A (X) is at least as large as in B (Y).</li>
<li>The use of a helical rise works best if the angle is 360 degrees.</li>
<li>If the cell size in C (Z) is set to the same value as the helical rise, the structure can be repeated in C (Z) to continue the helix structure.</li>
</ul>
</body></html>
"""
